import pathlib


def get_artifact_dir(hub, target_os="linux"):
    """
    function to get the full path to artifacts directory
    with the pkg_type included
    """
    artifacts_dir = pathlib.Path(hub.OPT.heist.artifacts_dir, "onedir", target_os)
    if not artifacts_dir.is_dir():
        artifacts_dir.mkdir(parents=True)
    return str(artifacts_dir)


def get_artifact_suffix(hub, target_os="linux"):
    """
    return the suffix for an artifact for
    each targeted OS.
    """
    suffix = "tar.xz"
    if target_os == "windows":
        suffix = ".zip"
    return suffix



async def sig_create(hub, name: str, target, **kwargs):
    pass


async def sig_send(hub, name: str, source: str, dest: str, **kwargs):
    pass


async def sig_get(hub, name: str, source: str, dest: str):
    pass


async def sig_cmd(hub, name: str, command: str, **kwargs):
    pass


async def sig_tunnel(hub, name: str, remote: str, local: str):
    pass


async def sig_destroy(hub, name: str):
    pass



import json
from typing import Any
from typing import Dict

import rend.exc


def generate_roster_data(hub, roster_data, roster_file=None):
    """
    Using json parse the data pass in by roster_data.
    If roster_file is also passed, write the roster_data
    to the given file.
    """
    try:
        ret = hub.rend.json.render(roster_data)
    except rend.exc.RenderException:
        hub.log.error("Invalid json data in roster_data")
        return False

    if roster_file:
        with open(roster_file, "w+") as fp_:
            json.dump(ret, fp_)
    return ret


async def read(
    hub, roster: str = None, roster_file: str = "", roster_data=None
) -> Dict[str, Any]:
    """
    Given the rosters to read in, the tgt and the tunnel_plugin
    """
    ret = {}

    ready = False
    if roster_data:
        ready = hub.roster.init.generate_roster_data(
            roster_data, roster_file=roster_file
        )

    # If a specific roster plugin wasn't specified then determine the best roster plugin automatically
    if roster is None:
        if roster_file.endswith(".fernet"):
            roster = "fernet"
        else:
            roster = "flat"
        hub.log.info(f"Picking default roster: {roster}")

    if not ready:
        ready = await hub.roster[roster].read(roster_file)

    if not ready:
        raise ValueError(f"The roster {roster} did not return data when rendered")

    if not isinstance(ready, dict):
        raise ValueError(f"The roster {roster} is not formatted correctly")

    for id_, condition in ready.items():
        if not isinstance(condition, dict):
            raise ValueError(f"The roster {roster} is not formatted correctly.")
        ret[condition.get("id", id_)] = condition
        if "id" not in condition:
            condition["id"] = id_

        ret[condition["id"]] = condition

    return ret



import os
from typing import Dict


def _get_defaults() -> Dict[str, str]:
    program_data = os.environ.get("ProgramData", "C:\\ProgramData")
    defaults = {
        "posix": {
            "config": "/etc/heist/heist.conf",
            "roster_dir": "/etc/heist/rosters",
            "roster_file": "/etc/heist/roster",
            "artifacts": "/var/tmp/heist/artifacts",
        },
        "nt": {
            "config": f"{program_data}\\heist\\heist.conf",
            "roster_dir": f"{program_data}\\heist\\rosters",
            "roster_file": f"{program_data}\\heist\\roster",
            "artifacts": f"{program_data}\\heist\\artifacts",
        },
    }.get(os.name, {})

    return defaults


OS_DEFAULTS = _get_defaults()

CLI_CONFIG = {
    "config": {"options": ["-c"], "subcommands": ["_global_"], "os": "HEIST_CONFIG"},
    "acct_profile": {"subcommands": ["_global_"]},
    "artifacts_dir": {"subcommands": ["_global_"]},
    "roster": {"subcommands": ["_global_"]},
    "roster_dir": {"subcommands": ["_global_"]},
    "roster_file": {"options": ["-R"], "subcommands": ["_global_"]},
    "roster_data": {"subcommands": ["_global_"]},
    "checkin_time": {"subcommands": ["_global_"]},
    "dynamic_upgrade": {"subcommands": ["_global_"]},
    "renderer": {"subcommands": ["_global_"]},
    "target": {"options": ["--tgt", "-t"], "subcommands": ["_global_"]},
    "artifact_version": {"options": ["-a, --artifact"], "subcommands": ["_global_"]},
    "service_plugin": {"options": ["-s", "--service"], "subcommands": ["_global_"]},
    "manage_service": {"options": ["--service-status"], "subcommands": ["_global_"]},
    "auto_service": {"subcommands": ["_global_"]},
    "clean": {
        "options": ["--clean"],
        "action": "store_true",
        "subcommands": ["_global_"],
    },
    "noclean": {
        "options": ["--noclean"],
        "action": "store_true",
        "subcommands": ["_global_"],
    },
    # ACCT options
    "acct_file": {"source": "acct", "os": "ACCT_FILE", "subcommands": ["_global_"]},
    "acct_key": {"source": "acct", "os": "ACCT_KEY", "subcommands": ["_global_"]},
}
CONFIG = {
    "config": {
        "default": OS_DEFAULTS.get("config"),
        "help": "Heist configuration location",
    },
    "acct_profile": {
        "default": "default",
        "help": "The specific named profile to read from encrypted acct files",
    },
    "artifacts_dir": {
        "default": OS_DEFAULTS.get("artifacts"),
        "help": "The location to look for artifacts that will be sent to target systems",
    },
    "roster": {
        "default": None,
        "help": "The type of roster to use to load up the remote systems to tunnel to",
    },
    "roster_dir": {
        "default": OS_DEFAULTS.get("roster_dir"),
        "help": "The directory to look for rosters",
    },
    "roster_file": {
        "options": ["-R"],
        "default": OS_DEFAULTS.get("roster_file"),
        "help": "Use a specific roster file, "
        "if this option is not used then the roster_dir will be used to find roster files",
    },
    "roster_data": {
        "default": None,
        "help": "Pass json data to be used for the roster data",
    },
    "checkin_time": {
        "default": 60,
        "type": int,
        "help": "The number of seconds between checking to see if the managed system needs to get an updated binary "
        "or agent restart.",
    },
    "dynamic_upgrade": {
        "default": False,
        "action": "store_true",
        "help": "Tell heist to detect when new binaries are available and dynamically upgrade target systems",
    },
    "renderer": {
        "default": "yaml",
        "help": "Specify the renderer to use to render heist roster files",
    },
    "target": {
        "options": ["--tgt", "-t"],
        "default": "",
        "help": "target used for multiple rosters",
    },
    "artifact_version": {
        "default": "",
        "help": "Version of the artifact to use for heist",
    },
    "roster_defaults": {
        "default": {},
        "type": dict,
        "help": "Default options to use for all rosters. CLI options will"
        "override these defaults",
    },
    "service_plugin": {
        "default": "raw",
        "help": "The type of service to use when managing the artifacts service status",
    },
    "auto_service": {
        "default": False,
        "type": bool,
        "help": "Auto detect the service manager to use on start up of service.",
    },
    "noclean": {
        "default": False,
        "action": "store_true",
        "help": "Whether to clean the deployed artifact and configurations",
    },
    "run_dir_root": {
        "default": False,
        "help": "Directory location on remote system for root deployment",
    },
}
SUBCOMMANDS = {
    # The manager determines how you want to create the tunnels and if you want to deploy
    # ephemeral agents to the remote systems
    "test": {x: "" for x in ("help", "desc")}
}
DYNE = {
    "acct": ["acct"],
    "artifact": ["artifact"],
    "heist": ["heist"],
    "roster": ["roster"],
    "service": ["service"],
    "tunnel": ["tunnel"],
    "tool": ["tool"],
}



import pkg_resources


def pre_get(hub, ctx):
    kwargs = ctx.get_arguments()
    version = kwargs.get("version")
    repo_data = kwargs.get("repo_data")
    if version:
        check_ver = version
        if "-" in version:
            check_ver, pkg_ver = version.split("-")
            assert isinstance(int(pkg_ver), int)
        valid_version = pkg_resources.safe_version(check_ver)
        assert check_ver == valid_version, f"version {check_ver} is not valid"



"""
Path operations for heist
"""
import os
import pathlib
import sys
from pathlib import Path
from pathlib import PurePosixPath
from pathlib import PureWindowsPath


def path_convert(hub, target_system, entry_path, ext_path):
    """
    Convert local path to target (remote) path.

    Due to pathlib.Path docstring indicating that a PureWindowsPath and
    PosixPath can not be created on their counterpart's system, we must
    make our own conversion.
    :param target_system:
    :param entry_path:
    :param list options:
    :return:
    """
    path_obj = None
    if "win" in target_system:
        path_obj = PureWindowsPath(entry_path)
        for option in ext_path:
            path_obj = path_obj.joinpath(option)
    elif "linux" in target_system or "darwin" in target_system:
        path_obj = PurePosixPath(entry_path)
        for option in ext_path:
            path_obj = path_obj.joinpath(option)
    hub.log.debug(f"Converted path for {target_system} for {entry_path}")
    return path_obj


def _realpath_darwin(path):
    base = ""
    for part in path.parts[1:]:
        if base != "":
            if os.path.islink(os.path.sep.join([base, part])):
                base = os.readlink(os.path.sep.join([base, part]))
            else:
                base = os.path.abspath(os.path.sep.join([base, part]))
        else:
            base = os.path.abspath(os.path.sep.join([base, part]))
    return base


def _realpath_windows(path):
    base = ""
    if not isinstance(path, Path):
        path = Path(path)
    for part in path.parts:
        if base != "":
            try:
                part = os.readlink(os.path.sep.join([base, part]))
                base = os.path.abspath(part)
            except OSError:
                base = os.path.abspath(os.path.sep.join([base, part]))
        else:
            base = part
    return base


def _realpath(path):
    """
    Cross platform realpath method. On Windows when python 3, this method
    uses the os.readlink method to resolve any filesystem links.
    All other platforms and version use ``os.path.realpath``.
    """
    if sys.platform == "darwin":
        return _realpath_darwin(path)
    elif sys.platform == "win32":
        return _realpath_windows(path)
    return os.path.realpath(path)


def clean_path(hub, root, path, subdir=False):
    """
    Return a clean path that has been validated.
    Using os.path here instead of pathlib, because
    the api's functionalities are not entirely the
    same yet.
    """
    real_root = _realpath(root)
    path = os.path.expandvars(path)
    if not os.path.isabs(real_root):
        return ""
    if not os.path.isabs(path):
        path = os.path.join(root, path)
    path = os.path.normpath(path)
    real_path = _realpath(path)
    if path.startswith(real_root):
        return os.path.join(root, path)
    if subdir:
        if real_path.startswith(real_root):
            return real_path
    else:
        if os.path.dirname(real_path) == os.path.normpath(real_root):
            return real_path
    return ""



import hashlib
import os
import pathlib
import re
import shutil
import tarfile
import tempfile
import warnings
import zipfile

import aiohttp.client_exceptions


def __init__(hub):
    hub.artifact.ACCT = ["artifact"]


async def extract(
    hub,
    target_name,
    tunnel_plugin,
    tmpdir=None,
    binary=None,
    run_dir=None,
    target_os="linux",
):
    """
    Extract the binary on the target.
    Uses tar for linux and unzip for windows.

    .. versionchanged:: v6.3.0
    """
    path = run_dir / pathlib.Path(binary).name

    if not hub.tool.path.clean_path(run_dir, pathlib.Path(binary).name):
        hub.log.error(f"The run_dir directory {run_dir} is not a valid path.")
        return False

    if tmpdir:
        warnings.warn(
            "The kwarg tmpdir is no longer used and will be removed in version v6.4.0"
        )
    if target_os == "linux":
        cmd = f"tar -xvf {path} --directory={run_dir}"
    elif target_os == "windows":
        cmd = f'powershell -command "$ProgressPreference = "SilentlyContinue"; Expand-Archive -LiteralPath "{path}" -DestinationPath {run_dir}"'
    ret = await hub.tunnel[tunnel_plugin].cmd(target_name, cmd, target_os=target_os)
    if ret.returncode == 0:
        return True
    hub.log.error(f"Could not extract {path} on the target")
    return False


def verify(hub, location, hash_value, hash_type="sha256"):
    with open(location, "rb") as fp:
        file_hash = getattr(hashlib, hash_type)(fp.read()).hexdigest()
        if not file_hash == hash_value:
            return False
    return True


async def fetch(hub, session, url, download=False, location=False):
    """
    Fetch a url and return json. If downloading artifact
    return the download location.
    """
    async with session.request("GET", url) as resp:
        if resp.status == 200:
            if download:
                with open(location, "wb") as fn_:
                    fn_.write(await resp.read())
                return location
            try:
                return await resp.json()
            except aiohttp.client_exceptions.ContentTypeError:
                return await resp.text()
        hub.log.critical(f"Cannot query url {url}. Returncode {resp.status} returned")
        return False


async def get(
    hub,
    artifact_name: str,
    target_os: str,
    version: str = None,
    repo_data=None,
    artifacts_dir=None,
    **kwargs,
):
    """
    Fetch a url return the download location.
    """
    if artifacts_dir is None:
        artifacts_dir = pathlib.Path(hub.OPT.heist.artifacts_dir)

    async with aiohttp.ClientSession() as session:
        with tempfile.TemporaryDirectory() as tmpdirname:
            # Download and verify the designated Heist manager artifact
            tmp_artifact_location = await hub.artifact[artifact_name].get(
                target_os=target_os,
                version=version,
                repo_data=repo_data,
                session=session,
                tmpdirname=pathlib.Path(tmpdirname),
                **kwargs,
            )
            if not tmp_artifact_location:
                return False

            artifact_location = artifacts_dir / tmp_artifact_location.name
            if not hub.tool.path.clean_path(artifacts_dir, tmp_artifact_location.name):
                hub.log.error(
                    f"The tmp artifact {tmp_artifact_location.name} is not in the correct directory"
                )
                return False

            # artifact is already downloaded we do not need to copy/check
            if tmp_artifact_location == artifact_location:
                return True

            hub.log.info(
                f"Copying the artifact {artifact_location.name} to {str(artifacts_dir)}"
            )
            shutil.move(tmp_artifact_location, artifact_location)

    # ensure artifact was downloaded
    if not any(str(version) in x for x in os.listdir(str(artifacts_dir))):
        hub.log.critical(
            f"Did not find the {version} artifact in {str(artifacts_dir)}."
            f" Untarring the artifact failed or did not include version"
        )
        return False
    return artifact_location


def version(hub):
    # TODO Determine which artifact to use, find the right plugin, and find out the target's version of the artifact
    ...


async def deploy(
    hub,
    target_name,
    tunnel_plugin,
    run_dir,
    binary,
    artifact_name,
    target_os="linux",
    run_dir_root=None,
    **kwargs,
):
    """
    Deploy the artifact
    """
    if not run_dir_root:
        run_dir_root = hub.heist.init.default(target_os, "run_dir_root")
    if not hub.tool.path.clean_path(run_dir_root, run_dir):
        hub.log.error(f"The run_dir path {run_dir} is not valid")
        return False

    cmd = "\n".join(
        [
            s.strip()
            for s in f"""/bin/sh << 'EOF'
if [ -d {run_dir.parent} ]; then
    ls {run_dir.parent};
    exit 0
else
    echo "The {run_dir.parent} does not exist. Deploying artifact"
    exit 1
fi
EOF""".split(
                "\n"
            )
        ]
    )

    if target_os == "windows":
        cmd = (
            "powershell -command "
            + '"'
            + "; ".join(
                [
                    f"$Folder = '{run_dir.parent}'",
                    "if (Test-Path -Path $Folder) {Get-ChildItem -name $Folder} else {exit 1}",
                ]
            )
            + '"'
        )

    # check if there was a previous heist deployment
    use_prev = False
    prev_artifact = None
    prev = await hub.tunnel[tunnel_plugin].cmd(target_name, cmd, target_os=target_os)
    if prev.returncode == 1:
        ret = await hub.artifact[artifact_name].deploy(
            target_name,
            tunnel_plugin,
            run_dir,
            binary,
            target_os=target_os,
            run_dir_root=run_dir_root,
            **kwargs,
        )
    else:
        run_dir = run_dir.parent / prev.stdout.strip()
        get_binary = f"ls {run_dir}/"
        use_prev = True
        ret = await hub.artifact[artifact_name].deploy(
            target_name,
            tunnel_plugin,
            run_dir,
            binary,
            run_dir_root=run_dir_root,
            target_os=target_os,
            verify=True,
            **kwargs,
        )
        artifact_files = await hub.tunnel[tunnel_plugin].cmd(
            target_name, get_binary, target_os=target_os
        )
        suffix = hub.tool.artifacts.get_artifact_suffix(target_os=target_os)
        prev_artifact = [x for x in artifact_files.stdout.split() if suffix in x]
        if prev_artifact:
            prev_artifact = prev_artifact[0]
    return run_dir, ret, use_prev, prev_artifact


async def verify_checksum(
    hub,
    target_name,
    tunnel_plugin,
    run_dir,
    source_fp,
    target_fp,
    target_os="linux",
    hash_type=None,
):
    """
    Verify checksum file
    """
    if hash_type is None:
        hash_type = "sha512"

    # validating hash_type
    if not getattr(hashlib, hash_type):
        hub.log.error(f"hash_type of {hash_type} is not valid")
        return False

    # validate paths
    if not hub.tool.path.clean_path(run_dir, target_fp):
        hub.log.error(f"Path is not valid: {target_fp}")
        return False
    if not pathlib.Path(source_fp).is_file():
        hub.log.error(f"Source path does not exist: {source_fp}")
        return False
    if not hub.tool.path.clean_path(
        hub.tool.artifacts.get_artifact_dir(target_os=target_os), source_fp
    ):
        hub.log.error(f"Source path is not valid: {source_fp}")
        return False

    if target_os == "windows":
        cmd = (
            "powershell -command "
            + '"'
            + "; ".join(
                [
                    f"Get-ChildItem -Recurse {target_fp.parent} | Get-FileHash -Algorithm {hash_type.upper()} | Format-List"
                ]
            )
            + '"'
        )
    else:
        cmd = f"cd {run_dir}; {hash_type}sum -c {target_fp}"
    ret = await hub.tunnel[tunnel_plugin].cmd(target_name, cmd, target_os=target_os)
    if ret.returncode != 0:
        hub.log.error(f"Could not verify the checksum of the artifact")
        return False
    if target_os == "windows":
        target_data = {}
        for line in ret.stdout.split(" : "):
            line = line.strip()
            if line.startswith(("Algorithm", "SHA")):
                continue
            if "\nPath" in line:
                _hash = "".join(line.split()).replace("Path", "")
                continue
            if "\nAlgorithm" in line:
                _path = line.split()[0]
            target_data[_hash] = _path
    else:
        cmd = f"cat {target_fp}"
        ret = await hub.tunnel[tunnel_plugin].cmd(target_name, cmd, target_os=target_os)

    with open(source_fp) as fp:
        data = fp.read()
        if target_os == "windows":
            items = iter(data.split())
            source_data = dict(zip(items, items))
            test = {}
            for key, value in source_data.items():
                key = key.upper()
                hash_len = getattr(hashlib, hash_type)().block_size
                check = r"([a-fA-F\d]){128}"
                is_hash = re.match(check, key)
                if is_hash:
                    _hash = key
                    _path = value
                else:
                    is_hash = re.match(check, value)
                    if is_hash:
                        _hash = value.upper()
                        _path = key

                test[_hash] = _path

                if not target_data.get(_hash):
                    hub.log.error(
                        f"Could not verify the checksum of the artifact for {_hash} and {_path}"
                    )
                    return False
                else:
                    target_data[_hash] == _path

        else:
            if data != ret.stdout:
                hub.log.error(f"Could not verify the checksum of the artifact")
                return False
    return ret


def checksum(hub, files, hash_type=None):
    """
    Create checksum for files to be copied over
    for the artifact.
    """
    if not hash_type:
        hash_type = "sha512"

    # validating hash_type
    if not getattr(hashlib, hash_type):
        hub.log.error(f"hash_type of {hash_type} is not valid")
        return False

    ret = {}
    for fp in files:
        fp = pathlib.Path(fp)
        if not fp.exists():
            hub.log.error(f"Could not create checksum. Path {fp} does not exist")
            return False
        hash_obj = getattr(hashlib, hash_type)()
        count = 1
        if fp.is_file():
            hash_obj.update(open(fp, "rb").read())
            ret[hash_obj.hexdigest()] = fp
        if fp.is_dir():
            for _fp in fp.rglob("*"):
                hash_obj = getattr(hashlib, hash_type)()
                count += 1
                if _fp.is_file():
                    hash_obj.update(open(_fp, "rb").read())
                    ret[hash_obj.hexdigest()] = _fp
    return ret


async def clean(hub, target_name, tunnel_plugin):
    """
    Clean up the deployed artifact and files
    """
    # remove run directory
    run_dir = hub.heist.CONS[target_name]["run_dir"]
    target_os, _ = await hub.tool.system.os_arch(target_name, tunnel_plugin)

    # clean alias symlinks
    if target_os == "linux":
        scripts_dir = run_dir / "scripts"
        cmd_alias = f"ls {scripts_dir}"
        get_files = await hub.tunnel[tunnel_plugin].cmd(
            target_name, cmd_alias, target_os=target_os
        )
        if get_files.returncode != 0:
            hub.log.error("Did not find alias files. Not removing")
        else:
            for _file in get_files.stdout.split():
                _file = str(scripts_dir / _file)
                find_symlink = f"find -L /usr/bin/ -samefile {_file}"
                get_symlink = await hub.tunnel[tunnel_plugin].cmd(
                    target_name, find_symlink, target_os=target_os
                )
                if get_symlink.returncode != 0:
                    hub.log.error(f"Could not find the symlink to script {_file}")
                    break
                rm_symlink = f"rm {get_symlink.stdout.strip()}"
                rm_symlink = await hub.tunnel[tunnel_plugin].cmd(
                    target_name, rm_symlink, target_os=target_os
                )
                if rm_symlink.returncode != 0:
                    hub.log.error(f"Could not remove the symlink of file {_file}")
                    break

    if target_os == "windows":
        cmd_isdir = f'powershell -command "If ( (Get-Item {run_dir}).PSIsContainer ) {{ exit 0 }} Else {{ exit 1 }}"'
        cmd_rmdir = f'powershell -command "cmd /c rmdir /s /q {run_dir}"'
    else:
        cmd_isdir = f"[ -d {run_dir} ]"
        cmd_rmdir = f"rm -rf {run_dir}"

    # Make sure it is a directory
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name, cmd_isdir, target_os=target_os
    )
    if ret.returncode == 0:
        await hub.tunnel[tunnel_plugin].cmd(target_name, cmd_rmdir, target_os=target_os)

    # remove parent directory if its empty
    # If it's not empty, there might be another running instance of heist that
    # was previously deployed
    await hub.tunnel[tunnel_plugin].cmd(
        target_name,
        f"rmdir {hub.heist.CONS[target_name]['run_dir'].parent}",
        target_os=target_os,
    )


async def create_aliases(hub, content, aliases=None, target_os="linux"):
    """
    Create the alias files that will be sent
    to the target
    """
    artifacts_dir = (
        pathlib.Path(hub.tool.artifacts.get_artifact_dir(target_os=target_os))
        / "scripts"
    )
    artifacts_dir.mkdir(exist_ok=True)
    # aliases should follow the dictionary structure:
    # {"binary-name": {"file": "path_to_binary_to_copy", "cmd": "cmd used to run on target"},
    #  "binary-name2": {"file": "path_to_binary_to_copy", "cmd": "cmd used to run on target"}}
    for alias in aliases:
        alias_file = aliases[alias]["file"]
        alias_cmd = aliases[alias]["cmd"]
        if target_os == "windows":
            alias_file = alias_file.parent / f"{alias_file.name}.bat"
        # this requires {alias} to be in the content.
        alias_content = content.format(alias=alias_cmd)
        with open(alias_file, "w") as fp:
            fp.write(alias_content)
        alias_file.chmod(0o711)
    return True


async def deploy_aliases(
    hub, target_name, tunnel_plugin, run_dir, aliases_dir, target_os="linux"
):
    """
    Deploy the aliases files to the target
    and setup correct path environment variables
    """
    artifacts_dir = (
        pathlib.Path(hub.tool.artifacts.get_artifact_dir(target_os=target_os))
        / "scripts"
    )
    await hub.tunnel[tunnel_plugin].send(
        target_name,
        artifacts_dir,
        run_dir,
        preserve=True,
        recurse=True,
    )
    if target_os == "windows":
        path_cmd = f'powershell -command "setx PATH "{aliases_dir};$env:PATH""'
        ret = await hub.tunnel[tunnel_plugin].cmd(
            target_name, path_cmd, target_os=target_os
        )
        if ret.returncode != 0:
            hub.log.error(f"Could not add {artifacts_dir} to windows PATH environment")
            return False
    else:
        path_cmd = f"ln -s {aliases_dir}/* /usr/bin/"
        ret = await hub.tunnel[tunnel_plugin].cmd(
            target_name, path_cmd, target_os=target_os
        )
        if ret.returncode != 0:
            hub.log.error(f"Could not add aliases to PATH")
            return False
    return True



import getpass
from typing import Dict

DEFAULT_TUNNEL = "asyncssh"


async def sig_run(
    hub,
    remotes: Dict[str, Dict[str, str]],
    artifact_version,
    manage_service,
    roster_file: str,
    roster: str,
    **kwargs
):
    ...


def _validate_remote(remote: Dict[str, str]):
    if not remote.get("tunnel"):
        remote["tunnel"] = DEFAULT_TUNNEL
    return remote


async def call_run(hub, ctx):
    kwargs = ctx.get_arguments()
    remotes = kwargs["remotes"]

    validate_remotes = {}
    for id_, remote in remotes.items():
        validate_remotes[id_] = _validate_remote(remote)

    kwargs.pop("remotes")
    return await ctx.func(
        kwargs.pop("hub"),
        remotes=validate_remotes,
        artifact_version=kwargs["artifact_version"],
        manage_service=kwargs["manage_service"],
        **kwargs["kwargs"]
    )



import inspect
import re
from typing import Any

import asyncssh
from dict_tools.data import NamespaceDict

__virtualname__ = "asyncssh"


def __init__(hub):
    """
    Set up the objects to hold connection instances
    """
    hub.tunnel.asyncssh.ACCT = ["tunnel.asyncssh"]
    hub.tunnel.asyncssh.CONS = {}


def _autodetect_asyncssh_opt(hub, option: str) -> Any:
    """"""
    # TODO autodetect certain options
    return None


def _get_asyncssh_opt(hub, target, option: str, default: Any = None) -> Any:
    """
    Get an assyncssh option from the target/roster first, if that fails get it from the config, if not there
    then try to autodetect the option (I.E. Checking for keys in the .ssh folder of the target)
    :param option:
    :param target:
    :return:
    """
    result = target.get(option)
    if not result and "heist" in hub.OPT:
        result = hub.OPT.heist.get(option)
        if not result:
            result = hub.OPT.heist.roster_defaults.get(option)
    if not result:
        result = _autodetect_asyncssh_opt(hub, option)
    if not result:
        result = default
    return result


async def create(hub, name: str, target, reconnect=False):
    """
    Create a connection to the remote system using a dict of values that map
    to this plugin. Name the connection for future use, the connection data
    will be stored on the hub under hub.tunnel.asyncssh.CONS
    :param name:
    :param target:
    """
    if name in hub.tunnel.asyncssh.CONS and not reconnect:
        hub.log.debug(f"tunnel {name} already exists")
        return True

    # The id MUST be in the target, everything else might be in the target, conf, or elsewhere
    target = target.copy()
    id_ = target.pop("host", None) or target["id"]

    possible_options = set(
        inspect.getfullargspec(asyncssh.SSHClientConnectionOptions.prepare).args
    )
    # Remove options from `inspect` that don't belong
    possible_options -= {"self", "args", "kwargs", "tunnel"}
    # Add connection options that aren't specified in `SSHClientConnectionOptions.prepare`
    possible_options.update(
        {"port", "loop", "family", "flags", "local_addr", "options"}
    )
    # Check for each possible SSHClientConnectionOption in the target, config, then autodetect (if necessary)
    con_opts = {"known_hosts": None}
    for arg in possible_options:
        opt = _get_asyncssh_opt(hub, target, arg)
        if opt is not None:
            con_opts[arg] = opt

    try:
        conn = await asyncssh.connect(id_, **con_opts)
    except Exception as e:
        hub.log.error(f"Failed to connect to '{id_}': {e}")
        return False
    sftp = await conn.start_sftp_client()

    hub.tunnel.asyncssh.CONS[name] = {
        "con": conn,
        "sftp": sftp,
        "sudo": target.get("sudo", None),
        "tty": target.get("tty"),
        "term_type": target.get("term_type"),
        "term_size": target.get("term_size"),
        "password": target.get("password"),
    }
    return True


async def send(hub, name: str, source: str, dest: str, preserve=False, recurse=False):
    """
    Take the file located at source and send it to the remote system

    :param name: The name of the asyncssh tunnel connection
    :param source: The full path to the source location.
    :param dest: The full path to the destination location.
    :param preserve: Preserve the permissions on the destination host for
         the files being copied.
    :param recurse: Recursively copy a directory over to the destination
         host
    """
    target = hub.tunnel.asyncssh.CONS[name]
    sftp = target["sftp"]
    if target["sudo"]:
        # TODO run SFTP with permissions
        pass
    await sftp.put(source, dest, preserve=preserve, recurse=recurse)


async def get(hub, name: str, source: str, dest: str):
    """
    Take the file located on the remote system and copy it locally
    """
    sftp = hub.tunnel.asyncssh.CONS[name]["sftp"]
    await sftp.get(source, dest)


async def cmd(
    hub,
    name: str,
    command: str,
    background=False,
    target_os="linux",
    sudo=None,
    **kwargs,
):
    """
    Execute the given command on the machine associated with the named connection
    """
    # Allow us to bypass sudo settings in roster for commands
    # that do not require sudo via kwargs
    if sudo is None:
        sudo = hub.tunnel.asyncssh.CONS[name].get("sudo")
    sudo_password = re.compile(r"(?:.*)[Pp]assword(?: for .*)?:", re.M)
    con: asyncssh.SSHClientConnection = hub.tunnel.asyncssh.CONS[name]["con"]
    if target_os == "windows":
        win_background_cmd = command
        arg_list = win_background_cmd.split()
        win_background_cmd = arg_list.pop(0)
        arg_list = " ".join(arg_list)
        win_background_cmd = (
            "powershell -command "
            + '"'
            + "; ".join(
                [
                    f"$action = New-ScheduledTaskAction -Execute '{win_background_cmd}' "
                    f"-Argument '{arg_list}'",
                    "$trigger = New-ScheduledTaskTrigger -Once -At 00:00",
                    "$principal = New-ScheduledTaskPrincipal "
                    '-UserId "$env:USERNAME" '
                    '-LogonType "S4U" '
                    "-RunLevel Highest",
                    "Register-ScheduledTask -Action $action "
                    "-Trigger $trigger "
                    "-Principal $principal "
                    "-TaskName heist-background "
                    "-Description 'Run background process' "
                    "-Force",
                    "Start-ScheduledTask -TaskName heist-background",
                    "Unregister-ScheduledTask -TaskName heist-background "
                    "-Confirm:$false",
                ]
            )
            + '"'
        )

    result = {}
    if hub.tunnel.asyncssh.CONS[name].get("tty"):
        kwargs["term_type"] = (
            hub.tunnel.asyncssh.CONS[name].get("term_type") or "xterm-color"
        )
        kwargs["term_size"] = hub.tunnel.asyncssh.CONS[name].get("term_size") or (
            80,
            24,
        )
    if sudo:
        if background:
            if target_os == "windows":
                command = win_background_cmd
            else:
                command = f"sudo -b su -c '{command} &'"
        else:
            if target_os != "windows":
                command = f"sudo {command}"
        async with con.create_process(command, **kwargs) as process:
            while not process.stdout.at_eof():
                output = await process.stdout.read(1024)
                if sudo_password.search(output):
                    process.stdin.write(
                        hub.tunnel.asyncssh.CONS[name].get("password") + "\n"
                    )
                stdout = output
                if not stdout:
                    stdout = await process.stdout.read()
                stderr = await process.stderr.readline()
                result = NamespaceDict(
                    stdout=stdout,
                    stderr=stderr,
                    returncode=process.returncode,
                    exit_status=process.returncode,
                )
    else:
        if background:
            if target_os == "windows":
                command = win_background_cmd
            else:
                command = f"{command} > /dev/null 2>&1 &"
        result = await con.run(command, **kwargs)
    return result


async def tunnel(hub, name: str, remote: str, local: str):
    """
    Given the local and remote addrs create a tcp tunnel through the connection
    """
    con = hub.tunnel.asyncssh.CONS[name]["con"]
    listener = await con.forward_remote_port("", remote, "localhost", local)


async def destroy(hub, name: str):
    """
    Destroy the named connection
    """
    con = hub.tunnel.asyncssh.CONS[name]["con"]
    con.close()
    await con.wait_closed()


def connected(hub, name: str):
    """
    Determine whether the connection is
    connected or disconnected.
    """
    return hub.tunnel.asyncssh.CONS[name]["con"]._transport



from typing import Any
from typing import Dict


async def read(hub, roster_file: str = "") -> Dict[str, Any]:
    """
    Read in the data from an encrypted roster
    """
    if roster_file:
        return hub.crypto.init.decrypt_file(
            roster_file, hub.OPT.acct.acct_key, crypto_plugin="fernet"
        )
    return {}



async def disable(hub, tunnel_plugin, target_name, service):
    await hub.service[hub.OPT.heist.service_plugin].disable(
        tunnel_plugin, target_name, service
    )


async def enable(hub, tunnel_plugin, target_name, service):
    await hub.service[hub.OPT.heist.service_plugin].enable(
        tunnel_plugin, target_name, service
    )


async def start(hub, tunnel_plugin, target_name, service, **kwargs):
    return await hub.service[hub.OPT.heist.service_plugin].start(
        tunnel_plugin, target_name, service, **kwargs
    )


async def stop(hub, tunnel_plugin, target_name, service, target_os="linux"):
    return await hub.service[hub.OPT.heist.service_plugin].stop(
        tunnel_plugin,
        target_name,
        service,
        target_os=target_os,
    )


async def restart(hub, target_name, tunnel_plugin, service, **kwargs):
    hub.log.debug(f"Restarting the service {service}")
    return await hub.service[hub.OPT.heist.service_plugin].restart(
        target_name, tunnel_plugin, service, **kwargs
    )


def get_service_plugin(hub, remote=None, grains=None):
    """
    Return the serivce manager plugin that should be used.
    """
    if not grains:
        grains = {}

    if not remote:
        remote = {}

    # Get plugin from Heist config first
    service_plugin = hub.OPT.heist.get("service_plugin")

    # Will be overwritten by auto_service, if configured
    if hub.OPT.heist.auto_service:
        if "init" in grains:
            hub.log.debug("Auto detecting the service manager")
            service_plugin = grains["init"]
        else:
            hub.log.error("Could not auto detect the service")

    # Will be overwritten by roster.cfg if defined
    if "service_plugin" in remote:
        service_plugin = remote["service_plugin"]
    return service_plugin


def service_conf_path(hub, service_name, service_plugin=None):
    if not service_plugin:
        service_plugin = hub.service.init.get_service_plugin()
    conf_path = hub.service[service_plugin].conf_path(service_name)
    if not conf_path:
        return ""
    return conf_path


async def clean(
    hub,
    target_name,
    tunnel_plugin,
    service_name,
    service_plugin=None,
    target_os="linux",
):
    if not service_plugin:
        service_plugin = hub.heist.CONS[target_name]["service_plugin"]
    # stop service
    await hub.service[service_plugin].stop(
        target_name, tunnel_plugin, service_name, target_os=target_os
    )

    service_conf = hub.service.init.service_conf_path(service_name)
    if service_conf:
        ret = await hub.tunnel[tunnel_plugin].cmd(target_name, f"[ -f {service_conf} ]")
        if ret.returncode == 0:
            await hub.tunnel[tunnel_plugin].cmd(target_name, f"rm -f {service_conf}")
        else:
            hub.log.error(
                f"The conf file {service_conf} does not exist. Will not attempt to remove"
            )

    await hub.service[service_plugin].clean(target_name, tunnel_plugin, service_name)


async def status(hub, target_name, tunnel_plugin, service, target_os="linux"):
    """
    Check the status of the service and return if it's
    running or not.
    """
    return await hub.service[hub.OPT.heist.service_plugin].status(
        target_name, tunnel_plugin, service
    )



from typing import Dict


async def run(hub, remotes: Dict[str, Dict[str, str]], artifact_version=None, **kwargs):
    hub.log.info(
        "This is a test heist manager. You have installed heist correctly. "
        "Install a heist manager to use full functionality"
    )
    return True



"""
The entry point for Heist, this is where config loading and the project
spine is set up
"""
import asyncio
import ipaddress
import pathlib
import socket
import subprocess
import sys
from typing import Any

from dict_tools.data import NamespaceDict


def __init__(hub):
    hub.heist.CONS = {}
    hub.heist.ROSTERS = {}

    # Load all relative subs onto the hub recursively
    hub.pop.sub.load_subdirs(hub.heist, recurse=True)
    for dyne in ("acct", "artifact", "rend", "roster", "service", "tunnel", "tool"):
        hub.pop.sub.add(dyne_name=dyne)
    hub.pop.sub.load_subdirs(hub.service, recurse=True)


def cli(hub):
    """
    Load the cli as the first thing we do to account for pluggable OS dependent systems.
    """
    hub.pop.config.load(["heist", "acct", "rend"], cli="heist")
    hub.heist.init.env()

    # Create a path on host system to cache downloads
    # This location is where grains and salt binary will be downloaded to
    pathlib.Path(hub.OPT.heist.artifacts_dir).mkdir(0o700, parents=True, exist_ok=True)
    if sys.platform.startswith("win"):
        subprocess.run(["icacls", hub.OPT.heist.artifacts_dir, "inheritance:d"])
        subprocess.run(
            ["icacls", hub.OPT.heist.artifacts_dir, "/remove", "Users", "/T"]
        )
        subprocess.run(["icacls", hub.OPT.heist.artifacts_dir, "/GRANT:R", "Users:(R)"])

    # If we are on windows, create a separate loop so we can handle the
    # keyboard interrupt. due to issues with current asyncio, windows, and
    # python together, we cannot support signals natively when running from
    # windows. therefore, we must diverge code here
    if "win" in sys.platform:
        hub.pop.loop.create()
        try:
            hub.pop.Loop.run_until_complete(hub.heist.init.start_win())
        except KeyboardInterrupt:
            hub.log.debug("Caught keyboard interrupt")
        finally:
            hub.pop.Loop.run_until_complete(hub.heist.init.clean())
            hub.pop.Loop.close()
    else:
        hub.heist.init.start_nix()


def env(hub):
    """
    Translate the config into the needed vars on the hub.
    :param hub:
    :return:
    """
    if hub.OPT["heist"].run_dir_root:
        run_dir_root = hub.OPT["heist"].run_dir_root
        hub.heist.OS_DEFAULTS = {
            "linux": {
                "os_path": "/var/tmp/heist",
                "user": "root",
                "run_dir_root": run_dir_root,
            },
            "windows": {
                "os_path": "C:\\Windows\\temp\\heist",
                "user": "Administrator",
                "run_dir_root": run_dir_root,
            },
            "default": {
                "os_path": "/var/tmp/heist",
                "user": "root",
                "run_dir_root": run_dir_root,
            },
        }
    else:
        hub.heist.OS_DEFAULTS = {
            "linux": {
                "os_path": "/var/tmp/heist",
                "user": "root",
                "run_dir_root": "/var/tmp/",
            },
            "windows": {
                "os_path": "C:\\Windows\\temp\\heist",
                "user": "Administrator",
                "run_dir_root": "C:\\Windows\\temp\\",
            },
            "default": {
                "os_path": "/var/tmp/heist",
                "user": "root",
                "run_dir_root": "/var/tmp/",
            },
        }


def default(hub, target_os: str, key: str) -> Any:
    os_defaults = hub.heist.OS_DEFAULTS.get(target_os)
    if not os_defaults:
        value = None
        hub.log.error(f"OS '{target_os}' not defined in hub.heist.OS_DEFAULTS")
    else:
        value = os_defaults.get(key)

    if not value:
        hub.log.error(f"No {target_os}-specific default for {key}, using fallback")
        value = hub.heist.OS_DEFAULTS["default"].get(key)

    return value


def start_nix(hub):
    """
    Start the async loop and get the process rolling for linux based systems
    """
    if not hub.SUBPARSER:
        hub.log.critical("Must set a heist manager")
        return False

    try:
        hub.pop.loop.start(
            hub.heist.init.run_remotes(
                hub.SUBPARSER,
                artifact_version=hub.OPT.heist.artifact_version,
                roster_file=hub.OPT.heist.roster_file,
                roster=hub.OPT.heist.roster,
                roster_data=hub.OPT.heist.roster_data,
                manage_service=hub.OPT.heist.manage_service,
                clean=hub.OPT.heist.clean,
            ),
            sigint=hub.heist.init.clean,
            sigterm=hub.heist.init.clean,
        )
    except asyncio.CancelledError:
        hub.log.debug("Cancelled remaining running asyncio tasks")
    finally:
        hub.pop.Loop.close()


async def start_win(hub):
    """
    Start the async loop and get the process rolling for nt based systems
    """
    if not hub.SUBPARSER:
        hub.log.critical("Must set a heist manager")
        return False

    await hub.heist.init.run_remotes(
        hub.SUBPARSER,
        artifact_version=hub.OPT.heist.artifact_version,
        roster_file=hub.OPT.heist.roster_file,
        roster=hub.OPT.heist.roster,
        roster_data=hub.OPT.heist.roster_data,
        manage_service=hub.OPT.heist.manage_service,
        clean=hub.OPT.heist.clean,
    )


def display_output(hub, out):
    """
    print results to the screen
    for the user
    """
    lines = []
    for target in out:
        for key, value in target.items():
            if target["retvalue"] == 0:
                # print green
                lines.append(f"\033[92m {key}: {value}\033[00m\n")
            else:
                # print red
                lines.append(f"\033[91m {key}: {value}\033[00m\n")
    print("".join(lines))


async def run_remotes(
    hub,
    manager: str,
    artifact_version=None,
    roster_file: str = "",
    roster: str = None,
    roster_data=None,
    manage_service=None,
    clean=False,
    **kwargs,
):
    """
    Configs, rosters and targets have been loaded, time to execute the
    remote system calls
    """
    try:
        raw_remotes = await hub.roster.init.read(
            roster, roster_file=roster_file, roster_data=roster_data
        )
    except ValueError as err:
        hub.log.error(err.args[0])
        return False
    remotes = NamespaceDict()
    for k, v in raw_remotes.items():
        remotes[k] = NamespaceDict(v)
    if not remotes:
        return False

    hub.log.debug(f"Heist manager is '{manager}'")

    if manage_service:
        valid_opts = ["start", "restart", "stop", "enable", "disable", "status"]
        if manage_service not in valid_opts:
            hub.log.error(f"The service cannot be managed with {manage_service}")
            hub.log.error(f"Please use one of {valid_opts}")
            return False
    ret = hub.heist[manager].run(
        remotes,
        artifact_version=artifact_version,
        manage_service=manage_service,
        clean=clean,
        **kwargs,
    )
    if asyncio.iscoroutine(ret):
        _ret = await ret

    hub.heist.init.display_output(out=_ret)
    return _ret


async def clean(hub, signal: int = None):
    """
    Clean up the connections
    """
    if signal:
        hub.log.warning(f"Got signal {signal}! Cleaning up connections")
    coros = []
    # First clean up the remote systems
    if hub.OPT.heist.noclean:
        hub.log.info("`noclean` was set. Will not clean up the minion")
    else:
        for target_name, r_vals in hub.heist.ROSTERS.items():
            if not r_vals.get("bootstrap"):
                vals = hub.heist.CONS[target_name]
                manager = vals["manager"]
                coros.append(
                    hub.heist[manager].clean(
                        target_name,
                        vals["tunnel_plugin"],
                        service_plugin=vals.get("service_plugin"),
                        vals=vals,
                    )
                )
        await asyncio.gather(*coros)
    # Then shut down connections
    coros = []
    for target_name, vals in hub.heist.CONS.items():
        tunnel_plugin = vals["tunnel_plugin"]
        coros.append(hub.tunnel[tunnel_plugin].destroy(target_name))
    await asyncio.gather(*coros)
    if sys.version_info >= (3, 7):
        tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
    else:
        tasks = [
            t for t in asyncio.Task.all_tasks() if t is not asyncio.Task.current_task()
        ]
    for task in tasks:
        task.cancel()


def ip_is_loopback(hub, addr):
    """
    helper function to determine if an addr
    or hostname is a loopback address
    """
    try:
        info = socket.getaddrinfo(addr, 0)
    except socket.gaierror:
        hub.log.critical("Could not determine if addr is loopback")
        return False
    return ipaddress.ip_address(info[0][-1][0]).is_loopback



import os
import pathlib
from typing import Any
from typing import Dict


async def read(hub, roster_file: str = "") -> Dict[str, Any]:
    """
    Read in the data from the configured rosters
    """
    ret = {}
    rend = hub.OPT.heist.renderer

    if pathlib.Path(roster_file).is_file():
        return await hub.rend.init.parse(roster_file, rend)
    else:
        hub.log.error(f"Missing roster file: {roster_file}")

    if pathlib.Path(hub.OPT.heist.roster_dir).is_dir():
        for fn in os.listdir(hub.OPT.heist.roster_dir):
            full = os.path.join(hub.OPT.heist.roster_dir, fn)
            ret.update(await hub.rend.init.parse(full, rend))
    return ret



import time


async def disable(hub, tunnel_plugin, target_name, service):
    pass


async def enable(hub, tunnel_plugin, target_name, service):
    pass


async def start(
    hub,
    target_name,
    tunnel_plugin,
    service,
    run_cmd=None,
    target_os="linux",
    run_dir=None,
    background=False,
    **kwargs,
):
    """
    Start the service in the background
    """
    # each heist manager will need to have a function
    # that validates the run_cmd input
    if not hub.tool.service.valid_run_cmd(run_cmd, run_dir, target_os=target_os):
        hub.log.error(
            f"The run command is not valid. Not starting the service {service}"
        )
        return False
    cmd = [f"{run_cmd}"]
    hub.log.info(f"Starting the service {service}")
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name, " ".join(cmd), background=background, target_os=target_os
    )
    if ret.returncode != 0:
        hub.log.error(ret.stderr)
        return False
    return True


async def stop(hub, target_name, tunnel_plugin, service, target_os="linux"):
    """
    Stop the service
    """
    if target_os == "windows":
        kill_cmd = (
            f'powershell -command "Stop-Process -Name {service} -ErrorAction SilentlyContinue; '
            'If ($?) { exit 0 } else { exit 1 }"'
        )
    else:
        kill_cmd = f"pkill -f {service}"
    hub.log.debug(f"Attempting to kill {service} with: {kill_cmd}")
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name, kill_cmd, target_os=target_os
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    hub.log.info(f"Successfully killed {service}")
    return True


async def restart(
    hub,
    target_name,
    tunnel_plugin,
    service,
    run_cmd=None,
    target_os="linux",
    run_dir=None,
    **kwargs,
):
    if await hub.service.raw.status(
        target_name, tunnel_plugin, service, target_os=target_os
    ):
        if not await hub.service.raw.stop(
            target_name, tunnel_plugin, service, target_os=target_os
        ):
            hub.log.error(f"Could not stop the service {service}")
            return False
        time.sleep(2)
    while not hub.service.raw.status(
        target_name, tunnel_plugin, service, target_os=target_os
    ):
        time.sleep(5)
        continue
    if not await hub.service.raw.start(
        target_name,
        tunnel_plugin,
        service,
        run_cmd=run_cmd,
        target_os=target_os,
        run_dir=run_dir,
    ):
        hub.log.error(f"Could not start the service {service}")
        return False
    return True


def conf_path(hub, service_name):
    pass


async def clean(hub, target_name, tunnel_plugin, service):
    pass


async def status(hub, target_name, tunnel_plugin, service, target_os="linux", **kwargs):
    cmd = [f"pgrep -f {service}"]
    if target_os == "windows":
        cmd = [f'powershell -command "get-process {service}"']
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name, " ".join(cmd), target_os=target_os, sudo=kwargs.get("sudo")
    )
    if ret.returncode != 0:
        hub.log.info(f"The service {service} is not running")
        return False
    hub.log.info(f"The service {service} is running")
    return True



"""
Scan a netmask or ipaddr for open ssh ports
"""
import asyncio
import ipaddress
import socket
import sys
from typing import Any
from typing import Dict


async def read(hub, roster_file: str = "") -> Dict[str, Any]:
    """
    read data from connection to specified ports
    to detect SSH ports
    """
    ret = {}
    target = hub.OPT.heist.get("target")
    if not target:
        hub.log.critical("Need to define a target for the scan roster")
        return ret
    ports = hub.OPT.heist.ssh_scan_ports
    if not isinstance(ports, list):
        # Comma-separate list of integers
        ports = list(map(int, str(ports).split(",")))
    try:
        addrs = [ipaddress.ip_address(target)]
    except ValueError:
        try:
            addrs = ipaddress.ip_network(hub.OPT.heist.target).hosts()
        except ValueError:
            pass

    for addr in addrs:
        for port in ports:
            hub.log.debug(f"Scanning host: {addr} on port: {port}")
            try:
                if addr.version == 4:
                    fam = socket.AF_INET
                elif addr.version == 6:
                    fam = socket.AF_INET6

                addr = addr.exploded
                reader, writer = await asyncio.open_connection(addr, port, family=fam)
                data = await reader.read(100)
                if "ssh" not in data.decode().lower():
                    hub.log.critical(
                        f"Connection successful to {addr} "
                        f"but SSH information not returned."
                        f"Port {port} might not be an SSH port."
                    )
                ret[addr] = {"host": addr, "port": port}
                writer.close()
                if sys.version_info >= (3, 7):
                    await writer.wait_closed()
            except ConnectionRefusedError:
                hub.log.critical(f"Not able to connect to host: {addr} on port: {port}")
    return ret



def sig_disable(hub, tunnel_plugin, target_name, service):
    ...


def sig_enable(hub, tunnel_plugin, target_name, service):
    ...


def sig_start(hub, target_name, tunnel_plugin, service, **kwargs):
    ...


def sig_stop(hub, target_name, tunnel_plugin, service, target_os):
    ...


def sig_restart(hub, target_name, tunnel_plugin, service, **kwargs):
    ...


def sig_clean(hub, target_name, tunnel_plugin, service):
    ...


def sig_status(hub, target_name, tunnel_plugin, service, **kwargs):
    ...


def pre(hub, ctx):
    kwargs = ctx.get_arguments()
    if ctx.func.__name__ in ["clean", "get_service_plugin", "valid_service_names"]:
        return True
    service_name = kwargs.get("service")
    if ctx.func.__name__ in ["service_conf_path", "clean", "conf_path"]:
        service_name = kwargs.get("service_name")
    valid_names = hub.service.init.valid_service_names()
    if service_name not in valid_names:
        raise ValueError(
            f"The service name {service_name} is not a valid name. "
            f"Please choose from the following: {valid_names}"
        )



from typing import Any
from typing import Dict
from typing import Mapping


def sig_read(hub, roster_file: str) -> Dict[str, Any]:
    ...


async def post_read(hub, ctx):
    ret = ctx.ret
    for data in ret.values():
        assert isinstance(data, Mapping)
    return ret



from typing import Tuple


async def os_arch(hub, target_name: str, tunnel_plugin: str) -> Tuple[str, str]:
    """
    Query the system for the OS and architecture type
    """
    DELIM = "'|'"
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name,
        f'echo "$OSTYPE{DELIM}$MACHTYPE{DELIM}$env:PROCESSOR_ARCHITECTURE"'
        # target_name, f'echo "$OSTYPE|$MACHTYPE|$env:PROCESSOR_ARCHITECTURE"'
    )
    assert not ret.returncode, ret.stderr
    kernel, arch, winarch = ret.stdout.lower().split("|", maxsplit=2)

    # Set the architecture bit
    if "64" in winarch or "64" in arch:
        os_arch = "amd64"
    else:
        os_arch = "i386"

    # Set the kernel bit
    if "linux" in kernel:
        kernel = "linux"
    elif "darwin" in kernel:
        kernel = "darwin"
    elif "bsd" in kernel:
        kernel = "bsd"
    elif winarch:
        kernel = "windows"
    else:
        raise ValueError(
            f"Could not determine arch from kernel: {kernel} arch: {arch} winarch: {winarch}"
        )
    hub.log.debug(f'Detected arch "{os_arch}" on target')
    return kernel, os_arch



#!/usr/bin/env python3
import os
import shutil

from setuptools import Command
from setuptools import setup

NAME = "heist"
DESC = "Ephemeral software tunneling and delivery system"

# Version info -- read without importing
_locals = {}
with open(f"{NAME}/version.py") as fp:
    exec(fp.read(), None, _locals)
VERSION = _locals["version"]
SETUP_DIRNAME = os.path.dirname(__file__)
if not SETUP_DIRNAME:
    SETUP_DIRNAME = os.getcwd()


with open("requirements/base.txt") as f:
    REQUIREMENTS = f.read().splitlines()

with open("README.rst", encoding="utf-8") as f:
    LONG_DESC = f.read()


class Clean(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for subdir in (NAME, "tests"):
            for root, dirs, files in os.walk(
                os.path.join(os.path.dirname(__file__), subdir)
            ):
                for dir_ in dirs:
                    if dir_ == "__pycache__":
                        shutil.rmtree(os.path.join(root, dir_))


def discover_packages():
    modules = []
    for root, _, files in os.walk(os.path.join(SETUP_DIRNAME, NAME)):
        pdir = os.path.relpath(root, SETUP_DIRNAME)
        modname = pdir.replace(os.sep, ".")
        modules.append(modname)
    return modules


setup(
    name=NAME,
    author="",
    author_email="",
    url="",
    version=VERSION,
    description=DESC,
    long_description=LONG_DESC,
    long_description_content_type="text/x-rst",
    python_requires=">=3.6",
    requirements_extra={"salt": ["heist-salt"]},
    install_requires=REQUIREMENTS,
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 5 - Production/Stable",
    ],
    entry_points={
        "console_scripts": [
            "heist = heist.scripts:start",
        ],
    },
    packages=discover_packages(),
    cmdclass={"clean": Clean},
)



import os


async def disable(hub, tunnel_plugin, target_name, service):
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name, f"systemctl disable {service}"
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


async def enable(hub, tunnel_plugin, target_name, service):
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name, f"systemctl enable {service}"
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


async def start(hub, target_name, tunnel_plugin, service, block=True, **kwargs):
    cmd = [f"systemctl start {service}"]
    if not block:
        cmd.append("--no-block")
    ret = await hub.tunnel[tunnel_plugin].cmd(target_name, " ".join(cmd))
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


async def stop(hub, target_name, tunnel_plugin, service, target_os="linux"):
    ret = await hub.tunnel[tunnel_plugin].cmd(target_name, f"systemctl stop {service}")
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


async def restart(hub, target_name, tunnel_plugin, service):
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name, f"systemctl restart {service}"
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


def conf_path(hub, service_name):
    return hub.tool.path.path_convert(
        "linux", "/etc", ["systemd", "system", f"{service_name}.service"]
    )


async def clean(hub, target_name, tunnel_plugin, service):
    await hub.tunnel[tunnel_plugin].cmd(target_name, f"systemctl daemon-reload")


async def status(hub, target_name, tunnel_plugin, service, **kwargs):
    cmd = [f"systemctl status {service}"]
    ret = await hub.tunnel[tunnel_plugin].cmd(target_name, " ".join(cmd))
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True



async def disable(hub, tunnel_plugin, target_name, service):
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name,
        f"Set-Service -Name {service} -StartupType Disabled; "
        f"If ($?) {{ exit 0 }} else {{ exit 1 }}",
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


async def enable(hub, tunnel_plugin, target_name, service):
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name,
        f"Set-Service -Name {service} -StartupType Automatic; "
        f"If ($?) {{ exit 0 }} else {{ exit 1 }}",
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


async def start(hub, target_name, tunnel_plugin, service, **kwargs):
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name,
        f"Start-Service -Name {service}; If ($?) {{ exit 0 }} else {{ exit 1 }}",
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


async def stop(hub, target_name, tunnel_plugin, service):
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name,
        f"Stop-Service -Name {service}; If ($?) {{ exit 0 }} else {{ exit 1 }}",
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


async def restart(hub, target_name, tunnel_plugin, service):
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name,
        f"Restart-Service -Name {service}; If ($?) {{ exit 0 }} else {{ exit 1 }}",
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


def conf_path(hub, service_name):
    pass


async def clean(hub, target_name, tunnel_plugin, service):
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name, f"cmd /c sc delete {service}"
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True


async def status(hub, target_name, tunnel_plugin, service, **kwargs):
    """
    Get status of a service on windows
    """
    ret = await hub.tunnel[tunnel_plugin].cmd(
        target_name,
        f"Get-Service {service}",
    )
    if ret.exit_status != 0:
        hub.log.error(ret.stderr)
        return False
    return True
